use ed25519_dalek::Signer;
use icepick_module::Module;
use serde::{Deserialize, Serialize};

#[derive(Serialize, Deserialize, Debug)]
#[serde(tag = "operation", content = "values", rename_all = "kebab-case")]
pub enum Operation {
    GetPubkey {},

    Sign { message: Vec<u8> },
}

#[derive(Serialize, Deserialize, Debug)]
pub struct Request {
    derived_keys: Option<Vec<[u8; 32]>>,

    #[serde(flatten)]
    operation: Operation,
}

#[derive(thiserror::Error, Debug)]
pub enum Error {}

pub struct Ed25519;

impl Module for Ed25519 {
    type Error = Error;

    type Request = Request;

    fn describe_operations() -> Vec<icepick_module::help::Operation> {
        use icepick_module::help::*;

        let message = Argument::builder()
            .name("message")
            .description("The message to sign, as an array of bytes.")
            .r#type(ArgumentType::Required)
            .build();

        let get_pubkey = Operation::builder()
            .name("get-pubkey")
            .description("Get an Ed25519 public key from the provided private key.")
            .build();

        let sign = Operation::builder()
            .name("sign")
            .description("Sign a message using an Ed25519 private key.")
            .build()
            .argument(&message);

        vec![get_pubkey, sign]
    }

    fn handle_request(request: Self::Request) -> Result<serde_json::Value, Self::Error> {
        let Request {
            derived_keys,
            operation,
        } = request;

        match operation {
            Operation::GetPubkey {} => {
                let key = derived_keys
                    .iter()
                    .flatten()
                    .next()
                    .map(ed25519_dalek::SigningKey::from_bytes)
                    .unwrap();
                let key = key.verifying_key().to_bytes();
                Ok(serde_json::json!({
                    "blob": {
                        "pubkey": key,
                    }
                }))
            }
            Operation::Sign { message } => {
                let key = derived_keys
                    .iter()
                    .flatten()
                    .next()
                    .map(ed25519_dalek::SigningKey::from_bytes)
                    .unwrap();
                let signature = key.sign(&message);
                Ok(serde_json::json!({
                    "blob": {
                        "signature": signature.to_vec(),
                    }
                }))
            }
        }
    }
}
